import { TypeScriptConfig } from './TypeScriptConfig';
import * as ts from 'typescript';
import { VFS } from './VFS';
import { DEFAULT_LIB_LOCATION } from './staticFiles';

export class JourneyLanguageServiceHost implements ts.LanguageServiceHost {
  public readonly fs: VFS;
  private readonly compilerOptions: ts.CompilerOptions;
  public readonly config: TypeScriptConfig;
  private readonly customTransformers: ts.CustomTransformers | undefined;
  public readonly defaultLibLocation: string;

  constructor(fs: VFS, config: TypeScriptConfig, customTransformers?: ts.CustomTransformers | undefined) {
    this.fs = fs;
    this.compilerOptions = config.compilerOptions;
    this.config = config;
    this.customTransformers = customTransformers;
    this.defaultLibLocation = config.defaultLibLocation ?? DEFAULT_LIB_LOCATION;
  }

  // --- language service host ---------------

  getCompilationSettings(): ts.CompilerOptions {
    return this.compilerOptions;
  }

  getScriptFileNames(): string[] {
    return this.fs.listFiles().filter((path) => {
      const info = this.fs.info(path);
      return info?.isScript;
    });
  }

  getScriptVersion(fileName: string): string {
    const info = this.fs.info(fileName);
    return info?.version ?? '';
  }

  getScriptSnapshot(fileName: string): ts.IScriptSnapshot | undefined {
    const node = this.fs.read(fileName);
    if (node == null) {
      return undefined;
    }
    return ts.ScriptSnapshot.fromString(node.content);
  }

  readFile = (fileName: string): string | undefined => {
    return this.fs.read(fileName)?.content ?? undefined;
  };

  fileExists = (fileName: string) => {
    return this.fs.info(fileName) != null;
  };

  private getModuleFileNames() {
    return this.fs.listFiles().filter((path) => {
      const info = this.fs.info(path);
      return info?.isModule;
    });
  }

  readDirectory = (
    path: string,
    extensions?: readonly string[],
    exclude?: readonly string[],
    include?: readonly string[],
    depth?: number
  ): string[] => {
    // Used for auto-completing import paths
    if (path == '') {
      // Avoid returning anything here
      return [];
    }
    const files = this.getModuleFileNames().filter((filename) => {
      if (!filename.startsWith(path)) {
        return false;
      }
      const rest = filename.substring(path.length);
      const fileDepth = rest.split('/').length;
      if (fileDepth > (depth ?? 1)) {
        return false;
      }
      if (extensions) {
        const matched = extensions.filter((ext) => filename.endsWith(ext));
        if (matched.length == 0) {
          return false;
        }
      }
      return true;
    });

    return files;
  };

  getDirectories = (directoryName: string): string[] => {
    // Used for auto-completing import paths
    if (directoryName == '') {
      // Special case to avoid returning "file:"
      return [];
    }
    const scripts = this.getModuleFileNames().filter((filename) => filename.startsWith(directoryName));
    let dirs: Record<string, boolean> = {};
    for (let script of scripts) {
      const rest = script.substring(directoryName.length);
      const components = rest.split('/');
      if (components.length > 1 && components[0] != '') {
        dirs[components[0]] = true;
      }
    }
    return Object.keys(dirs);
  };

  getScriptKind?(fileName: string): ts.ScriptKind {
    const suffix = fileName.substr(fileName.lastIndexOf('.') + 1);
    switch (suffix) {
      case 'ts':
        return ts.ScriptKind.TS;
      case 'tsx':
        return ts.ScriptKind.TSX;
      case 'js':
        return ts.ScriptKind.JS;
      case 'jsx':
        return ts.ScriptKind.JSX;
      default:
        return this.getCompilationSettings().allowJs ? ts.ScriptKind.JS : ts.ScriptKind.TS;
    }
  }

  getCurrentDirectory(): string {
    return '';
  }

  getDefaultLibFileName(options: ts.CompilerOptions): string {
    return `${this.defaultLibLocation}${ts.getDefaultLibFileName(options)}`;
  }

  resolveTypeReferenceDirectives(
    typeDirectiveNames: string[],
    containingFile: string,
    redirectedReference: ts.ResolvedProjectReference | undefined,
    options: ts.CompilerOptions
  ): (ts.ResolvedTypeReferenceDirective | undefined)[] {
    let resolved: (ts.ResolvedTypeReferenceDirective | undefined)[] = [];
    const resolutionHost = this;
    for (let directive of typeDirectiveNames) {
      const result = ts.resolveTypeReferenceDirective(
        directive,
        containingFile,
        options,
        resolutionHost,
        redirectedReference
      );
      resolved.push(result?.resolvedTypeReferenceDirective);
    }
    return resolved;
  }

  resolveModuleNames(
    moduleNames: string[],
    containingFile: string,
    reusedNames: string[] | undefined,
    redirectedReference: ts.ResolvedProjectReference | undefined,
    options: ts.CompilerOptions
  ): (ts.ResolvedModule | undefined)[] {
    const resolvedModules: (ts.ResolvedModule | undefined)[] = [];
    const resolutionHost = this;
    for (const moduleName of moduleNames) {
      // try to use standard resolution
      const result = ts.resolveModuleName(
        moduleName,
        containingFile,
        this.compilerOptions,
        resolutionHost,
        undefined,
        redirectedReference
      );
      if (result.resolvedModule) {
        resolvedModules.push(result.resolvedModule);
        continue;
      } else if (moduleName.startsWith('.')) {
        // Relative path, consider not found - an error
      } else {
        if (this.config.requireHack) {
          // Try with classic
          const classicResult = ts.classicNameResolver(
            moduleName,
            containingFile,
            this.compilerOptions,
            resolutionHost
          );
          if (classicResult.resolvedModule) {
            resolvedModules.push(classicResult.resolvedModule);
            continue;
          }
        }
        // The module could not be found.
        // These generate TS2307 errors, which we then filter out before presenting errors to the user.
      }

      resolvedModules.push(undefined);
    }
    return resolvedModules;
  }

  getCustomTransformers(): ts.CustomTransformers | undefined {
    return this.customTransformers;
  }

  mapDefinitions<T extends ts.DocumentSpan>(defs: readonly T[] | null): (T | null)[] {
    return (
      defs
        ?.map((def) => {
          return this.mapDefinition(def);
        })
        .filter((def) => def != null) ?? []
    );
  }

  mapDefinition<T extends ts.DocumentSpan>(def: T): T | null {
    if (def == null) {
      return null;
    }
    const path = def.fileName;
    const info = this.fs.info(path);
    if (info == null || !info.generated) {
      // Direct source file
      return def;
    }
    const node = this.fs.read(path);
    if (node == null) {
      // Cannot find file
      return null;
    }
    const mapped = node.mapTargetToSource(def.textSpan);
    if (mapped) {
      return {
        ...def,
        fileName: mapped.path,
        textSpan: {
          start: mapped.start,
          length: mapped.length
        },
        originalFileName: def.fileName,
        originalTextSpan: def.textSpan
      };
    } else {
      // Cannot find this specific definition in a source file
      return null;
    }
  }
}
