import { SyncFileSystem } from '@journeyapps/source-utils';
import { SingleTransform } from './SingleTransform';
import { applicationParser } from '@journeyapps/core';
import { Schema } from '@journeyapps/parser-schema';
import { Version } from '@journeyapps/parser-common';
import { schemaToDefinitions } from './schemaToDefinitions';
import { cachedSingleGenerate, getInputsVersions } from './caching';

export interface SchemaTransformOptions {
  dbs: string[];
  asynchronous: boolean;
  includeUser?: boolean;
  definitionPath: string;
  sourcePath: string;
  applicationXmlPath: string;
}

export class SchemaTransform extends SingleTransform {
  private options: SchemaTransformOptions;
  private schemaCache: Schema | null = null;
  private fs: SyncFileSystem;

  constructor(fs: SyncFileSystem, options: SchemaTransformOptions) {
    super({
      inputs: [options.sourcePath, options.applicationXmlPath],
      target: options.definitionPath,
      isScript: true
    });
    this.fs = fs;
    this.options = options;
  }

  loadSchema(): Schema | null {
    // Populate cache
    this.generateSingle();

    return this.schemaCache;
  }

  getVersion() {
    return getInputsVersions(this.fs, this.inputs);
  }

  generateSingle = cachedSingleGenerate({
    version: () => this.getVersion(),
    generate: () => {
      const text = this.fs.readTextSync(this.options.sourcePath);
      const applicationXml = this.fs.readTextSync(this.options.applicationXmlPath);
      const appInfo = applicationParser.parse(applicationXml);
      const schema = new Schema();
      schema.loadXml(text, { apiVersion: new Version('4.0'), recordSource: true });
      this.schemaCache = schema;
      return schemaToDefinitions(schema, {
        sourcePath: this.options.sourcePath,
        asynchronous: this.options.asynchronous,
        dbs: this.options.dbs,
        mobileUserType: this.options.includeUser ? appInfo.mobileUserType : undefined
      });
    }
  });
}
